/* gvSIG Mini. A free mobile phone viewer of free maps.
 *
 * Copyright (C) 2010 Prodevelop.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,USA.
 *
 * For more information, contact:
 *
 *   Prodevelop, S.L.
 *   Pza. Don Juan de Villarrasa, 14 - 5
 *   46001 Valencia
 *   Spain
 *
 *   +34 963 510 612
 *   +34 963 510 968
 *   prode@prodevelop.es
 *   http://www.prodevelop.es
 *
 *   gvSIG Mini has been partially funded by IMPIVA (Instituto de la Peque�a y
 *   Mediana Empresa de la Comunidad Valenciana) &
 *   European Union FEDER funds.
 *   
 *   2010.
 *   author Alberto Romeu aromeu@prodevelop.es
 *   
 */

package es.prodevelop.android.spatialindex.quadtree.provide;

import java.util.Collection;

import es.prodevelop.android.spatialindex.poi.Metadata;
import es.prodevelop.gvsig.mini.exceptions.BaseException;
import es.prodevelop.gvsig.mini.geom.Extent;
import es.prodevelop.gvsig.mini.utiles.Cancellable;

/**
 * Implementations of this interface should be able to query the quadtree (its
 * persisted version) to retrieve POIs
 * 
 * @author aromeu
 * 
 */
public interface QuadtreeProvider {

	/**
	 * Search for POIs on the Quadtree contained into a bounding box
	 * 
	 * @param boundingBox
	 *            The bounding box to search into
	 * @param zoomLevel
	 *            The current zoom level of the layer
	 * @return A Collection of POI or its subclasses
	 * @throws BaseException
	 *             If an unexpected error occurs
	 */
	public Collection getPOIs(Extent boundingBox, int zoomLevel,
			Cancellable cancellable) throws BaseException;

	/**
	 * Search for POIs giving a name/description
	 * 
	 * @param name
	 *            The name/description of the POI to search
	 * @param exactMatch
	 *            If false makes a 'LIKE' search
	 * @return
	 * @throws BaseException
	 */
	public Collection getPOIsByName(String name, boolean exactMatch,
			Cancellable cancellable) throws BaseException;

	public Collection getPOIs(Extent boundingBox, double res, int zoomLevel,
			Cancellable cancellable) throws BaseException;
	
	public Metadata getPOIMetadata();
	
	public Metadata getStreetMetadata();
	
	public String getDatabasePath();

}
