/* gvSIG Mini. A free mobile phone viewer of free maps.
 *
 * Copyright (C) 2010 Prodevelop.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,USA.
 *
 * For more information, contact:
 *
 *   Prodevelop, S.L.
 *   Pza. Don Juan de Villarrasa, 14 - 5
 *   46001 Valencia
 *   Spain
 *
 *   +34 963 510 612
 *   +34 963 510 968
 *   prode@prodevelop.es
 *   http://www.prodevelop.es
 *
 *   gvSIG Mini has been partially funded by IMPIVA (Instituto de la Peque�a y
 *   Mediana Empresa de la Comunidad Valenciana) &
 *   European Union FEDER funds.
 *   
 *   2010.
 *   author Alberto Romeu aromeu@prodevelop.es
 *   
 */

package es.prodevelop.android.spatialindex.quadtree.provide.fs;

import java.io.DataInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.util.Collection;

import es.prodevelop.android.spatialindex.poi.Metadata;
import es.prodevelop.android.spatialindex.quadtree.QuadtreeConstants;
import es.prodevelop.android.spatialindex.quadtree.memory.Tree;
import es.prodevelop.android.spatialindex.quadtree.memory.TreeLeaf;
import es.prodevelop.android.spatialindex.quadtree.memory.TreeNode;
import es.prodevelop.android.spatialindex.quadtree.provide.QuadtreeProvider;
import es.prodevelop.gvsig.mini.exceptions.BaseException;
import es.prodevelop.gvsig.mini.geom.Extent;
import es.prodevelop.gvsig.mini.utiles.Cancellable;
import es.prodevelop.gvsig.mini.utiles.Constants;

/**
 * A Provider for a BucketPRQuadtree persisted using FSQuadtreePersister.
 * 
 * @see Tree
 * 
 * @author aromeu
 * 
 */
public class FSQuadtreeProvider implements QuadtreeProvider {

	private Tree tree;

	private String pathOfIndexFile;

	/**
	 * The constructor
	 * 
	 * @param pathOfIndexFile
	 *            The path where the quadtree index file is stored. The name of
	 *            the index file is taken from
	 *            es.prodevelop.android.spatialindex
	 *            .quadtree.persist.fs#FSQuadtreePersister#INDEX_FILE_NAME
	 */
	public FSQuadtreeProvider(String pathOfIndexFile) {
		this.pathOfIndexFile = pathOfIndexFile + File.separator
				+ QuadtreeConstants.INDEX_FILE_NAME;
	}

	/**
	 * @see Tree#getPOIs(Extent)
	 */
	public Collection getPOIs(Extent boundingBox, int zoomLevel, Cancellable cancellable) throws BaseException {
		return getIndex().getPOIs(boundingBox);
	}

	private Tree getIndex() throws BaseException {
		if (tree == null) {
			this.loadIndex();
			return tree;
		} else {
			return tree;
		}
	}

	private void loadIndex() throws BaseException {

		File f = null;
		DataInputStream is = null;
		FileInputStream indexFile = null;

		final String fileName = pathOfIndexFile;
		f = new File(fileName);

		TreeNode[] parentNodes;
		int numChild = 0;
		try {
			if (f.exists()) {
				indexFile = new FileInputStream(fileName);
				is = new DataInputStream(indexFile);

				double minX = is.readDouble();
				double minY = is.readDouble();
				double maxX = is.readDouble();
				double maxY = is.readDouble();

				Extent maxExtent = new Extent(minX, minY, maxX, maxY);

				TreeNode root = processNode(is);
				if (root == null) {
					return;
				}
				if (root instanceof TreeLeaf) {
					return;
				}
				parentNodes = new TreeNode[1];
				parentNodes[0] = root;
				numChild += 4;
				processNodes(parentNodes, is, numChild);
				tree = new Tree(root, maxExtent);
			} else {
				throw new BaseException("ERROR: Index file doesn't exist!");
			}
		} catch (Exception e) {
			throw new BaseException(e);
		} finally {
			Constants.closeStream(indexFile);
			Constants.closeStream(is);
		}
	}

	private TreeNode processNode(DataInputStream b) {
		try {
			// short level = b.readShort();
			// System.out.println(""+numNodes++);
			// short node = b.readShort();
			short file = b.readShort();
			boolean hasBranch = b.readBoolean();

			if (hasBranch) {
				TreeNode res = new TreeNode();
				res.setHasBranch(hasBranch);
				return res;
			} else if (file != -1) {
				TreeLeaf res = new TreeLeaf();
				res.setArchive(file);
				return res;
			} else {
				// TreeNode res = new TreeNode();
				// res.hasBranch = hasBranch;
				// return res;
				return null;
			}
			// return null;
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
	}

	private void processNodes(TreeNode[] parentNodes, DataInputStream b,
			int numChild) {
		TreeNode[] tempNodes = new TreeNode[numChild];
		numChild = 0;
		int j = 0;

		try {

			if (b.available() <= 0) {
				return;
			}

			for (int i = 0; i < parentNodes.length; i++) {
				if (parentNodes[i] != null) {
					if (parentNodes[i].isHasBranch()) {
						TreeNode temp = processNode(b);
						if (temp != null) {
							if (temp.isHasBranch()) {
								numChild += 4;
							}
						}
						parentNodes[i].setUL(temp);
						tempNodes[j] = temp;
						j++;
						temp = processNode(b);
						if (temp != null) {
							if (temp.isHasBranch()) {
								numChild += 4;
							}
						}
						parentNodes[i].setUR(temp);
						tempNodes[j] = temp;
						j++;
						temp = processNode(b);
						if (temp != null) {
							if (temp.isHasBranch()) {
								numChild += 4;
							}
						}
						parentNodes[i].setBL(temp);
						tempNodes[j] = temp;
						j++;
						temp = processNode(b);
						if (temp != null) {
							if (temp.isHasBranch()) {
								numChild += 4;
							}
						}
						parentNodes[i].setBR(temp);
						tempNodes[j] = temp;
						j++;
					}
				}
			}

			processNodes(tempNodes, b, numChild);

		} catch (java.lang.OutOfMemoryError es) {
			es.printStackTrace();
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	public Collection getPOIsByName(String name, boolean exactMatch, Cancellable cancellable)
			throws BaseException {
		// FIXME: Implement this method
		throw new BaseException("Not implemented");
	}

	public Collection getPOIs(Extent boundingBox, double res, int zoomLevel, Cancellable cancellable)
			throws BaseException {
		// TODO Auto-generated method stub
		return null;
	}

	public Metadata getPOIMetadata() {
		// TODO Auto-generated method stub
		return null;
	}

	public Metadata getStreetMetadata() {
		// TODO Auto-generated method stub
		return null;
	}

	public String getDatabasePath() {
		return pathOfIndexFile;
	}
}
