/* gvSIG Mini. A free mobile phone viewer of free maps.
 *
 * Copyright (C) 2010 Prodevelop.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,USA.
 *
 * For more information, contact:
 *
 *   Prodevelop, S.L.
 *   Pza. Don Juan de Villarrasa, 14 - 5
 *   46001 Valencia
 *   Spain
 *
 *   +34 963 510 612
 *   +34 963 510 968
 *   prode@prodevelop.es
 *   http://www.prodevelop.es
 *
 *   gvSIG Mini has been partially funded by IMPIVA (Instituto de la Peque�a y
 *   Mediana Empresa de la Comunidad Valenciana) &
 *   European Union FEDER funds.
 *   
 *   2010.
 *   author Alberto Romeu aromeu@prodevelop.es
 */

package es.prodevelop.android.spatialindex.quadtree.provide.perst;

import java.util.ArrayList;
import java.util.Collection;

import es.prodevelop.android.spatialindex.poi.Metadata;
import es.prodevelop.android.spatialindex.poi.OsmPOI;
import es.prodevelop.android.spatialindex.poi.OsmPOIStreet;
import es.prodevelop.android.spatialindex.quadtree.persist.perst.POISpatialIndexRoot;
import es.prodevelop.android.spatialindex.quadtree.persist.perst.PerstDBHelper;
import es.prodevelop.android.spatialindex.quadtree.provide.BookmarkProviderListener;
import es.prodevelop.android.spatialindex.quadtree.provide.QuadtreeProvider;
import es.prodevelop.gvsig.mini.exceptions.BaseException;
import es.prodevelop.gvsig.mini.geom.Extent;
import es.prodevelop.gvsig.mini.utiles.Cancellable;
import es.prodevelop.gvsig.mini.utiles.WorkQueue;

public class PerstBookmarkProvider implements QuadtreeProvider {

	private String databasePath;
	private PerstDBHelper helper;
	private BookmarkProviderListener listener;

	public PerstBookmarkProvider(String databasePath) {
		this.databasePath = databasePath;
		helper = new PerstDBHelper();
		helper.openDatabase(this.databasePath, PerstDBHelper.DEFAULT_PAGE_POOL);
		POISpatialIndexRoot root = new POISpatialIndexRoot(helper.db);
		helper.initialize(root);
	}

	public void setQuadtreeProviderListener(BookmarkProviderListener listener) {
		this.listener = listener;
	}

	public void clearQuadtreeProviderListener() {
		this.listener = null;
	}

	public ArrayList getStreets() {
		return ((POISpatialIndexRoot) helper.getRoot()).getStreetIndex()
				.getPrefixList("");
	}

	public ArrayList getPOIs() {
		return ((POISpatialIndexRoot) helper.getRoot()).getNameFieldIndex()
				.getPrefixList("");
	}

	public void getPOIsAsynch() throws BaseException {
		if (this.listener == null)
			throw new BaseException("No QuadtreeProviderListener attached!");
		WorkQueue.getInstance().execute(new Runnable() {

			public void run() {
				ArrayList pois = getPOIs();
				listener.onPOISRetrieved(pois, true, null);
			}
		});
	}

	public void getStreetsAsynch() throws BaseException {
		if (this.listener == null)
			throw new BaseException("No QuadtreeProviderListener attached!");
		WorkQueue.getInstance().execute(new Runnable() {

			public void run() {
				ArrayList pois = getStreets();
				listener.onStreetsRetrieved(pois, true, null);
			}
		});
	}

	public boolean addStreet(OsmPOIStreet street) {
		String des = street.getDescription();
		street.setDescription(street.getDescription()
				+ street.getBoundingBox().toPrintString());
		boolean res = ((POISpatialIndexRoot) helper.getRoot()).getStreetIndex()
				.put(street);
		street.setDescription(des);
		helper.commit();
		if (listener != null)
			listener.onStreetAdded(street);
		return res;

	}

	public boolean addPOI(OsmPOI poi) {
		String des = poi.getDescription();
		poi.setDescription(poi.getDescription() + poi.toShortString(4));
		boolean res = ((POISpatialIndexRoot) helper.getRoot())
				.getNameFieldIndex().put(poi);
		poi.setDescription(des);
		helper.commit();
		if (listener != null)
			listener.onPOIAdded(poi);
		return res;
	}

	public boolean removeStreet(OsmPOIStreet street) {
		String des = street.getDescription();
		street.setDescription(street.getDescription()
				+ street.getBoundingBox().toPrintString());
		// helper.getRoot()).getStreetIndex().contains(street);
		boolean res = ((POISpatialIndexRoot) helper.getRoot()).getStreetIndex()
				.remove(street);
		street.setDescription(des);
		helper.commit();
		if (listener != null)
			listener.onStreetRemoved(street);
		return res;
	}

	public boolean removePOI(OsmPOI poi) {
		String des = poi.getDescription();
		poi.setDescription(poi.getDescription() + poi.toShortString(4));
		boolean res = ((POISpatialIndexRoot) helper.getRoot())
				.getNameFieldIndex().remove(poi);
		poi.setDescription(des);
		helper.commit();
		if (listener != null)
			listener.onPOIRemoved(poi);
		return res;
	}

	public PerstDBHelper getHelper() {
		return helper;
	}

	public Collection getPOIs(Extent boundingBox, int zoomLevel,
			Cancellable cancellable) throws BaseException {
		throw new BaseException("Stub!");
	}

	public Collection getPOIsByName(String name, boolean exactMatch,
			Cancellable cancellable) throws BaseException {
		throw new BaseException("Stub!");
	}

	public Collection getPOIs(Extent boundingBox, double res, int zoomLevel,
			Cancellable cancellable) throws BaseException {
		throw new BaseException("Stub!");
	}

	public Metadata getPOIMetadata() {
		// TODO Auto-generated method stub
		return null;
	}

	public Metadata getStreetMetadata() {
		// TODO Auto-generated method stub
		return null;
	}

	public String getDatabasePath() {
		return this.getDatabasePath();
	}
}
