/* gvSIG Mini. A free mobile phone viewer of free maps.
 *
 * Copyright (C) 2010 Prodevelop.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,USA.
 *
 * For more information, contact:
 *
 *   Prodevelop, S.L.
 *   Pza. Don Juan de Villarrasa, 14 - 5
 *   46001 Valencia
 *   Spain
 *
 *   +34 963 510 612
 *   +34 963 510 968
 *   prode@prodevelop.es
 *   http://www.prodevelop.es
 *
 *   gvSIG Mini has been partially funded by IMPIVA (Instituto de la Peque�a y
 *   Mediana Empresa de la Comunidad Valenciana) &
 *   European Union FEDER funds.
 *   
 *   2010.
 *   author Alberto Romeu aromeu@prodevelop.es
 */

package es.prodevelop.android.spatialindex.quadtree.provide.perst;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Enumeration;
import java.util.Hashtable;

import org.garret.perst.Key;
import org.garret.perst.RectangleR2;
import org.garret.perst.SpatialIndexR2;

import es.prodevelop.android.spatialindex.cluster.Cluster;
import es.prodevelop.android.spatialindex.poi.OsmPOI;
import es.prodevelop.android.spatialindex.poi.POICategories;
import es.prodevelop.android.spatialindex.quadtree.bucket.mr.MRBucketPRQuadtree;
import es.prodevelop.android.spatialindex.quadtree.memory.cluster.ClusterNode;
import es.prodevelop.android.spatialindex.quadtree.memory.cluster.ClusterTree;
import es.prodevelop.android.spatialindex.quadtree.persist.perst.ClusterTreePersistent;
import es.prodevelop.android.spatialindex.quadtree.persist.perst.OSMPoiClusterSpatialIndexRoot;
import es.prodevelop.android.spatialindex.quadtree.persist.perst.SpatialIndexRoot;
import es.prodevelop.android.spatialindex.quadtree.provide.LoadCategoryListener;
import es.prodevelop.android.spatialindex.quadtree.provide.QuadtreeProviderListener;
import es.prodevelop.gvsig.mini.common.impl.CollectionQuickSort;
import es.prodevelop.gvsig.mini.exceptions.BaseException;
import es.prodevelop.gvsig.mini.geom.Extent;
import es.prodevelop.gvsig.mini.geom.Point;
import es.prodevelop.gvsig.mini.map.ViewPort;
import es.prodevelop.gvsig.mini.utiles.Cancellable;
import es.prodevelop.gvsig.mini.utiles.Utilities;
import es.prodevelop.gvsig.mini.utiles.WorkQueue;

public class PerstOsmPOIClusterProvider extends PerstOsmPOIProvider {

	private int partsCounter = 0;
	private int totalParts;

	private Hashtable categoriesQuadtrees = new Hashtable();
	private ArrayList selectedCategories = new ArrayList();

	private int maxClusterLevel = 0;
	private int totalZoomLevels = 0;

	private QuadtreeProviderListener listener;

	private Extent extent = new Extent();

	private int currentZoomLevel;
	private int maxZoomLevel;

	private ArrayList clusters = new ArrayList();
	private ArrayList pois = new ArrayList();

	private Cancellable currentCancellable;
	private Extent maxExtent = new Extent(-180, -90, 180, 90);

	private Hashtable maxClusterSizeCat = new Hashtable();

	private final static int FEW_CATEGORIES_BBOX_FACTOR = 64;
	private final static int MANY_CATEGORIES_BBOX_FACTOR = 128;

	private final static int FEW_CATEGORIES_NUMBER = 5;

	public PerstOsmPOIClusterProvider(String databasePath, int totalZoomLevels,
			QuadtreeProviderListener listener, int maxZoomLevel)
			throws BaseException {
		super(databasePath);
		this.maxZoomLevel = maxZoomLevel;
		this.setTotalZoomLevels(totalZoomLevels);
		this.listener = listener;
		final int length = POICategories.ORDERED_CATEGORIES.length;

		// for (int i = 2; i < length; i++) {
		// this.addCategory(POICategories.ORDERED_CATEGORIES[i]);
		// this.addCategory(POICategories.PLACES);
		// this.addCategory(POICategories.TRANSPORTATION);
		// this.addCategory(POICategories.FOOD);
		// }

		// this.getClusterTree(POICategories.ALL);
	}

	/**
	 * Sets the categories of the provider, the categories previously loaded
	 * that are not in the categories param are removed from the provider.
	 * 
	 * This method also tries to load a ClusterTree per category, so it needs to
	 * access to the database, retrieve the persisted ClusterTree of each
	 * category and recursively load it. It can freeze the UI, better launch
	 * this method on a separate thread
	 * 
	 * @param categories
	 *            A list of categories @see POICategories
	 * @param listener
	 *            A LoadCategoryListener to receive notifications each time a
	 *            category is loaded. It can be null
	 */
	public void loadCategories(final ArrayList categories,
			final LoadCategoryListener listener) {
		WorkQueue.getPOIInstance().execute(new Runnable() {
			public void run() {
				if (categories == null)
					return;

				final int size = categories.size();

				Enumeration keys = PerstOsmPOIClusterProvider.this.categoriesQuadtrees
						.keys();

				String cat;
				boolean isOnList = false;

				// iterate the current list of categories of the provider
				while (keys.hasMoreElements()) {
					isOnList = false;
					cat = keys.nextElement().toString();
					// iterate the new list of categories
					for (int i = 0; i < size; i++) {
						// category is already in the list of categories of the
						// provider, do nothing
						if (categories.get(i).toString().compareTo(cat) == 0) {
							isOnList = true;
						}
					}
					// after iterate if the category is not in the list, remove
					// it from
					// the list of categories of the provider
					if (!isOnList) {
						PerstOsmPOIClusterProvider.this.removeCategory(cat,
								listener);
					}
				}

				// finally iterate the list of categories to add
				String category;
				for (int i = 0; i < size; i++) {
					category = categories.get(i).toString();
					try {
						PerstOsmPOIClusterProvider.this.addCategory(
								category.toLowerCase(), listener);
					} catch (BaseException e) {
						System.out.println(e.getMessage());
					}
				}
			}
		});

		// if (categories.size() == 0)
		// if (this.listener != null) {
		// listener.onClustersRetrieved(new ArrayList(), null, true, null);
		// listener.onPOISRetrieved(new ArrayList(), true, null);
		// }
	}

	/**
	 * Adds the category to the list of categories of the provider. If the
	 * category is already on the list does nothing.
	 * 
	 * This method calls to {@link #loadIndex(String)} so it is recommended to
	 * execute it on a background thread
	 * 
	 * @param category
	 * @see POICategories
	 * @throws BaseException
	 */
	public void addCategory(final String category,
			final LoadCategoryListener listener) throws BaseException {

		if (listener != null)
			listener.onLoadingCategoryIndex(category);
		partsCounter = 0;
		if (PerstOsmPOIClusterProvider.this.categoriesQuadtrees
				.containsKey(category)) {
			if (listener != null)
				listener.onCategoryIndexLoaded(category);
			return;
		}

		try {
			PerstOsmPOIClusterProvider.this.loadIndex(category);
		} catch (BaseException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

		// this.addSelectedCategory(category);

		if (listener != null)
			listener.onCategoryIndexLoaded(category);
	}

	public boolean isLoaded() {
		return (this.categoriesQuadtrees.size() > 0);
	}

	public void addSelectedCategory(String category) {
		if (this.categoriesQuadtrees.containsKey(category))
			this.selectedCategories.add(category);
	}

	public void removeSelectedCategory(String category) {
		this.selectedCategories.remove(category);
	}

	/**
	 * Removes a ClusterTree of the category given
	 * 
	 * @param category
	 * @see POICategories
	 */
	public void removeCategory(final String category,
			final LoadCategoryListener listener) {
		if (listener != null)
			listener.onRemovingCategoryIndex(category);
		if (!this.categoriesQuadtrees.containsKey(category)) {
			if (listener != null)
				listener.onCategoryIndexRemoved(category);
			return;
		}
		ClusterTree tree = (ClusterTree) this.categoriesQuadtrees
				.remove(category);
		tree = null;

		// this.removeSelectedCategory(category);

		if (listener != null)
			listener.onCategoryIndexRemoved(category);
	}

	protected Extent calcBufferExtent(Extent boundingBox) {
		final double width = boundingBox.getWidth() / 4;
		final double height = boundingBox.getHeight() / 4;

		Extent buffer = new Extent(boundingBox.getMinX() - width,
				boundingBox.getMinY() - height, boundingBox.getMaxX() + width,
				boundingBox.getMaxY() + height);

		return buffer;
	}

	protected Extent calcExtentToRetrieve(Extent boundingBox) {
		final double width = boundingBox.getWidth() / 6;
		final double height = boundingBox.getHeight() / 6;

		Extent buffer = new Extent(boundingBox.getMinX() - width,
				boundingBox.getMinY() - height, boundingBox.getMaxX() + width,
				boundingBox.getMaxY() + height);

		return buffer;
	}

	public Collection getPOIs(Extent boundingBox, int zoomLevel,
			Cancellable cancellable) throws BaseException {
		// if (extent.contains(boundingBox)) return null;
		this.setCurrentZoomLevel(zoomLevel);

		Extent extent = calcBufferExtent(boundingBox);

		final RectangleR2 rect = new RectangleR2(extent.getMinY(),
				extent.getMinX(), extent.getMaxY(), extent.getMaxX());
		final ArrayList pois = (ArrayList) getIndex().getList(rect);

		final int length = selectedCategories.size();
		if (length == POICategories.ORDERED_CATEGORIES.length)
			return pois;

		final int size = pois.size();

		OsmPOI p;
		ArrayList categoryPois = new ArrayList();

		boolean isValid = true;
		for (int i = 0; i < size; i++) {
			if (checkCancelled(cancellable, true)) {
				return this.pois;
			}
			isValid = false;
			p = (OsmPOI) pois.get(i);
			for (int j = 0; j < length; j++) {
				if (p.getCategory().compareTo(
						selectedCategories.get(j).toString()) == 0) {
					isValid = true;
					break;
				}
			}
			if (isValid)
				categoryPois.add(p);
		}

		this.extent = extent;
		return categoryPois;
	}

	public boolean shouldShowPOIS(final int zoomLevel) {
		return zoomLevel > (maxZoomLevel - 2);
	}

	public void getPOIsAsynch(final Extent boundingBox, final double res,
			final int zoomLevel, final Cancellable cancellable,
			final boolean forceGet) throws BaseException {
		try {
			if (this.selectedCategories.size() == 0) {
				if (listener != null)
					listener.onClustersRetrieved(null, null, true, cancellable);
				if (listener != null)
					listener.onPOISRetrieved(null, true, cancellable);
				return;
			}

			if (!maxExtent.contains(boundingBox))
				return;
			final Extent retrieveExtent = PerstOsmPOIClusterProvider.this
					.calcExtentToRetrieve(boundingBox);

			// if we change the zoom level, or go away from the buffer extent,
			// cancel the current task
			if ((zoomLevel == PerstOsmPOIClusterProvider.this
					.getCurrentZoomLevel() && !PerstOsmPOIClusterProvider.this.extent
					.contains(boundingBox))
					|| zoomLevel != PerstOsmPOIClusterProvider.this
							.getCurrentZoomLevel()) {
				if (currentCancellable != null)
					// System.out.println("1 cancelled");
					currentCancellable.setCanceled(true);
			}

			WorkQueue.getPOIInstance().clearPendingTasks();
			WorkQueue.getPOIInstance().execute(new Runnable() {
				public void run() {
					try {

						if (!forceGet
								&& zoomLevel == PerstOsmPOIClusterProvider.this
										.getCurrentZoomLevel()
								&& PerstOsmPOIClusterProvider.this.extent
										.contains(retrieveExtent)) {
							// System.out.println("2 cancelled");
							// currentCancellable.setCanceled(true);
							if (listener != null)
								listener.onClustersRetrieved(clusters, null,
										true, cancellable);
							if (listener != null)
								listener.onPOISRetrieved(pois, true,
										cancellable);
							return;
						}

						final Extent currentExtent = PerstOsmPOIClusterProvider.this.extent;
						final int currentZoomLevel = PerstOsmPOIClusterProvider.this.currentZoomLevel;
						// PerstOsmPOIClusterProvider.this.currentCancellable =
						// cancellable;

						long t1 = System.currentTimeMillis();

						int factor = FEW_CATEGORIES_BBOX_FACTOR;

						if (selectedCategories != null
								&& selectedCategories.size() > FEW_CATEGORIES_NUMBER) {
							factor = MANY_CATEGORIES_BBOX_FACTOR;
						}

						ArrayList clusters = (ArrayList) getPOIs(boundingBox,
								ViewPort.LAT_LON_RES[zoomLevel] * factor,
								zoomLevel, cancellable);
//						System.out.println("Time to get pois: "
//								+ (System.currentTimeMillis() - t1));

						t1 = System.currentTimeMillis();
						ArrayList pois = (ArrayList) getPOIsFromClusters(
								clusters, boundingBox, zoomLevel, cancellable);
//						System.out.println("Time to get clusters: "
//								+ (System.currentTimeMillis() - t1));

						// if (pois != null) {
						// DistanceToPointPOI sorter = new DistanceToPointPOI(
						// new Point(minXY[0], minXY[1]));
						// Object[] orderedPOIS = sorter.sort(pois);
						//
						// final int length = orderedPOIS.length;
						//
						// for (int i = 0; i < length; i++) {
						// pois.set(i, orderedPOIS[i]);
						// }
						// }

						// if (!isCanceled) {
						final Cancellable cancellable = PerstOsmPOIClusterProvider.this.currentCancellable;

						if (clusters != null)
							if (cancellable != null
									&& !cancellable.getCanceled())
								PerstOsmPOIClusterProvider.this.clusters = clusters;
						if (pois != null)
							if (cancellable != null
									&& !cancellable.getCanceled())
								PerstOsmPOIClusterProvider.this.pois = pois;

						if (cancellable != null && cancellable.getCanceled()) {
							// System.out.println("6");
							PerstOsmPOIClusterProvider.this.extent = currentExtent;
							PerstOsmPOIClusterProvider.this.currentZoomLevel = currentZoomLevel;
						}

						// }
					} catch (Exception e) {
						System.out.println(e.getMessage());
						// Log.e("", e.getMessage());
					}
				}
			});
		} catch (Exception e) {
			System.out.println(e.getMessage());
		}
	}

	private boolean checkCancelled(final Cancellable cancellable,
			final boolean pois) {
		if (cancellable != null && cancellable.getCanceled()) {
			if (listener != null)
				if (pois)
					listener.onPOISRetrieved(this.pois, true, cancellable);
				else
					listener.onClustersRetrieved(this.clusters, null, true,
							cancellable);
			return true;
		}
		return false;
	}

	public Collection getPOIs(Extent boundingBox, double res, int zoomLevel,
			Cancellable cancellable) throws BaseException {

		if (zoomLevel != this.getCurrentZoomLevel()) {
			this.clusters = new ArrayList();
			// System.out.println("3");
			if (listener != null)
				listener.onClustersRetrieved(clusters, null, true, cancellable);
		}

		// We get more clusters than fit the screen
		Extent extent = this.calcBufferExtent(boundingBox);
		this.extent = extent;
		this.currentZoomLevel = zoomLevel;
		this.currentCancellable = cancellable;

		Collection clusters = new ArrayList();

		final ArrayList categories = this.selectedCategories;
		final int length = categories.size();

		// At closer zoom levels, return no clusters (only pois)
		if (shouldShowPOIS(zoomLevel)) {
			if (listener != null)
				listener.onClustersRetrieved(clusters, null, true, cancellable);
			return null;
		} else {

			String cat;
			Collection catClusters;
			ClusterTree tree;
			for (int i = 0; i < length; i++) {
				cat = categories.get(i).toString();
				tree = this.getClusterTree(cat);
				if (tree != null) {
					long t1 = System.currentTimeMillis();
					catClusters = this.getClusterTree(cat)
							.getNodes(extent, res);
					maxClusterSizeCat.put(cat,
							new Integer(this.getClusterTree(cat)
									.getMaxClusterSize()));
//					System.out.println("Time to get nodes: "
//							+ (System.currentTimeMillis() - t1));
					if (checkCancelled(cancellable, false)) {
						// System.out.println("5");
						if (listener != null)
							listener.onClustersRetrieved(this.clusters, null,
									true, cancellable);
						return this.clusters;
					}
					// notify for each category
					t1 = System.currentTimeMillis();
					if (listener != null)
						listener.onClustersRetrieved(catClusters, cat, false,
								cancellable);
					clusters.addAll(catClusters);
//					System.out.println("Time to on clusters retrieved: "
//							+ (System.currentTimeMillis() - t1));
				}
			}

			// for (int i = levels; i <= levels; i++) //{
			// long t1 = System.currentTimeMillis();

			// Log.d("TIME",
			// "Time to memory quadtree: "
			// + (System.currentTimeMillis() - t1) + " ms");
			// Iterator it = clusters.iterator();
			//
			// t1 = System.currentTimeMillis();
			// while (it.hasNext()) {
			// c = (ClusterNode) it.next();
			// if (c != null) {
			// cl = (Cluster) root.getIDIndex().get(new Key(c.getID()));
			// if (cl.isValid())
			// clusterList.add(cl);
			// }
			// }
			// Log.d("TIME",
			// "Time to retrieve clusters: "
			// + (System.currentTimeMillis() - t1) + " ms");
			// }
		}

		this.extent = extent;

		// finally notify with all the clusters retrieved
		if (listener != null)
			listener.onClustersRetrieved(clusters, null, true, cancellable);

		return clusters;
	}

	public Cluster getClusterByID(int ID) {
		OSMPoiClusterSpatialIndexRoot root = (OSMPoiClusterSpatialIndexRoot) getHelper()
				.getRoot();

		Cluster cl = (Cluster) root.getIDIndex().get(new Key(ID));
		if (cl.isValid())
			return cl;
		return null;

	}

	public void expandClusterAsynch(final Cluster cluster,
			final QuadtreeProviderListener listener) {
		try {
			WorkQueue.getPOIInstance().execute(new Runnable() {
				public void run() {
					Cluster cl = getClusterByID(cluster.getID());
					final Extent extent = cl.getBoundingBox();
					if (cl != null) {
						final RectangleR2 rect = new RectangleR2(extent
								.getMinY(), extent.getMinX(), extent.getMaxY(),
								extent.getMaxX());
						final ArrayList pois = (ArrayList) getIndex().getList(
								rect);
						final String category = POICategories.ORDERED_CATEGORIES[cluster
								.getCat()];
						final int size = pois.size();

						OsmPOI poi;
						ArrayList poisOfCategory = new ArrayList();
						for (int i = 0; i < size; i++) {
							poi = (OsmPOI) pois.get(i);
							if (poi.getCategory().compareTo(category) == 0) {
								poisOfCategory.add(poi);
							}
						}

						boolean wasRemoved = clusters.remove(cluster);

						if (listener != null) {
							listener.onClusterExpanded(poisOfCategory, false,
									null, cluster);
							listener.onClustersRetrieved(clusters, "", true,
									null);
						}

					}
				}
			});
		} catch (Exception e) {
			if (listener != null)
				listener.onPOISRetrieved(null, false, null);
		}
	}

	public Collection getPOIsFromClusters(Collection clusters, Extent extent,
			int zoomLevel, Cancellable cancellable) {
		if (zoomLevel != this.getCurrentZoomLevel()) {
			this.pois = new ArrayList();
			if (listener != null)
				listener.onPOISRetrieved(pois, true, cancellable);
		}

		ArrayList pois = new ArrayList();

		if (clusters != null) {
			final int size = clusters.size();

			Cluster c;
			final ArrayList cl = (ArrayList) clusters;
			for (int i = 0; i < size; i++) {
				c = (Cluster) cl.get(i);
				if (c.isExpandable(extent)) {
					expandClusterAsynch(c, listener);
				}
			}
		}

		if (shouldShowPOIS(zoomLevel)) {
			try {
				ArrayList p = (ArrayList) getPOIs(extent, zoomLevel,
						cancellable);
				if (checkCancelled(cancellable, true)) {
					return this.pois;
				}
				if (listener != null)
					listener.onPOISRetrieved(p, false, cancellable);
				pois.addAll(p);
			} catch (BaseException e) {
				// TODO Auto-generated catch block
				// Log.e("", e.getMessage());
			}
		}

		if (listener != null)
			listener.onPOISRetrieved(pois, true, cancellable);

		return pois;
	}

	public void setCurrentZoomLevel(int currentZoomLevel) {
		this.currentZoomLevel = currentZoomLevel;
	}

	public int getCurrentZoomLevel() {
		return currentZoomLevel;
	}

	public Extent getExtent() {
		return extent;
	}

	SpatialIndexR2 getIndex() {
		return ((SpatialIndexRoot) getHelper().getRoot()).getPoiIndex();
	}

	public ClusterTree getClusterTree(final String category)
			throws BaseException {

		if (this.categoriesQuadtrees.containsKey(category)) {
			final ClusterTree tree = (ClusterTree) this.categoriesQuadtrees
					.get(category);
			return tree;
		} else
			return null;
	}

	private void loadIndex(final String category) throws BaseException {

		maxClusterLevel = 0;
		ClusterTreePersistent tree = (ClusterTreePersistent) ((OSMPoiClusterSpatialIndexRoot) getHelper()
				.getRoot()).getClusterTreeIndex().get(category);

		if (tree == null)
			throw new BaseException(
					"No cluster tree in the database for category: " + category);
		if (tree.getTree().compareTo("") == 0)
			throw new BaseException(
					"No cluster tree in the database for category: " + category);

		String clusterTreeString = tree.getTree();

		ClusterNode[] parentNodes;
		int numChild = 0;
		try {
			// get the extent
			String[] parts = clusterTreeString.split("#");

			// if (parts.length != 5)
			// throw new BaseException(
			// "Bad format of cluster tree in the database for category: "
			// + category);

			double minX = Double.valueOf(parts[0]).doubleValue();
			double minY = Double.valueOf(parts[1]).doubleValue();
			double maxX = Double.valueOf(parts[2]).doubleValue();
			double maxY = Double.valueOf(parts[3]).doubleValue();

			parts = parts[4].split("\\|");

			if (parts.length <= 0)
				throw new BaseException(
						"Bad format of cluster tree in the database for category: "
								+ category);

			Extent maxExtent = new Extent(minX, minY, maxX, maxY);

			totalParts = parts.length;

			ClusterNode root = processNode(parts, maxExtent);
			if (root == null) {
				return;
			}

			if (!root.isHasBranch())
				return;

			parentNodes = new ClusterNode[1];
			parentNodes[0] = root;
			numChild += 4;
			processNodes(parentNodes, parts, numChild);

			this.categoriesQuadtrees.put(category, new ClusterTree(root,
					maxExtent, maxClusterLevel));

		} catch (Exception e) {
			throw new BaseException(e);
		} finally {

		}
	}

	private ClusterNode processNode(String[] b, Extent rootExtent) {
		try {
			if (this.partsCounter >= totalParts) {
				return null;
			}

			final String part = b[this.partsCounter++];
			final String[] parts = part.split(";");
			final int ID = Integer.valueOf(parts[0]).intValue();
			final double x = Double.valueOf(parts[1]).doubleValue();
			final double y = Double.valueOf(parts[2]).doubleValue();
			final int numItems = Integer.valueOf(parts[3]).intValue();
			final boolean hasBranch = Boolean.valueOf(parts[4]).booleanValue();
			final short cat = Short.valueOf(parts[5]).shortValue();

			ClusterNode node = new ClusterNode();
			node.setBoundingBox(rootExtent);
			node.setID(ID);
			node.setX(x);
			node.setY(y);
			node.setNumItems(numItems);
			node.setHasBranch(hasBranch);
			node.setCat(cat);
			return node;
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
	}

	private void processNodes(ClusterNode[] parentNodes, String[] b,
			int numChild) {
		ClusterNode[] tempNodes = new ClusterNode[numChild];
		numChild = 0;
		int j = 0;

		try {

			if (this.partsCounter >= totalParts || parentNodes.length == 0) {
				return;
			}

			Extent bounds[];
			for (int i = 0; i < parentNodes.length; i++) {
				if (parentNodes[i] != null) {
					bounds = parentNodes[i].getExtents();
					if (parentNodes[i].isHasBranch()) {
						ClusterNode temp = processNode(b, bounds[0]);
						if (temp != null) {
							if (temp.isHasBranch()) {
								numChild += 4;
							}
						}
						parentNodes[i].setUL(temp);
						tempNodes[j] = temp;
						j++;
						temp = processNode(b, bounds[1]);
						if (temp != null) {
							if (temp.isHasBranch()) {
								numChild += 4;
							}
						}
						parentNodes[i].setUR(temp);
						tempNodes[j] = temp;
						j++;
						temp = processNode(b, bounds[2]);
						if (temp != null) {
							if (temp.isHasBranch()) {
								numChild += 4;
							}
						}
						parentNodes[i].setBL(temp);
						tempNodes[j] = temp;
						j++;
						temp = processNode(b, bounds[3]);
						if (temp != null) {
							if (temp.isHasBranch()) {
								numChild += 4;
							}
						}
						parentNodes[i].setBR(temp);
						tempNodes[j] = temp;
						j++;
					}
				}
			}
			maxClusterLevel++;

			processNodes(tempNodes, b, numChild);

		} catch (java.lang.OutOfMemoryError es) {
			es.printStackTrace();
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	/**
	 * @return the totalZoomLevels
	 */
	public int getTotalZoomLevels() {
		return totalZoomLevels;
	}

	/**
	 * @param totalZoomLevels
	 *            the totalZoomLevels to set
	 */
	public void setTotalZoomLevels(int totalZoomLevels) {
		this.totalZoomLevels = totalZoomLevels;
	}
	
	/**
	 * @param maxZoomLevel
	 *            the maxZoomlevel to set
	 */
	public void setMaxZoomLevel(int maxZoomLevel) {
		this.maxZoomLevel = maxZoomLevel;
	}

	public void setSelectedCategories(ArrayList categories)
			throws BaseException {
		this.selectedCategories.clear();
		this.selectedCategories.addAll(categories);
		if (categories.size() != 0) {
			Extent e = (Extent) this.extent.clone();
			this.extent = new Extent();
			int zoomLevel = this.currentZoomLevel;
			currentZoomLevel = 0;

			if (this.currentCancellable != null)
				this.currentCancellable.setCanceled(true);
			// FIXME res is not used at the moment, but take care if the API
			// changes, 0 can cause errors!!
			this.getPOIsAsynch(e, 0, zoomLevel, Utilities.getNewCancellable(),
					true);
		}
	}

	public ArrayList getSelectedCategories() {
		return selectedCategories;
	}

	private class DistanceToPointPOI extends CollectionQuickSort {

		private Point point;

		public DistanceToPointPOI(Point point) {
			this.point = point;
		}

		public boolean less(Object x, Object y) {
			double e1 = ((Point) x).distance(point);
			double e2 = ((Point) y).distance(point);

			return e1 < e2;
		}
	}

	/**
	 * @return the currentCancellable
	 */
	public Cancellable getCurrentCancellable() {
		return currentCancellable;
	}

	public Hashtable getMaxClusterSizeCat() {
		return maxClusterSizeCat;
	}
}
