/* gvSIG Mini. A free mobile phone viewer of free maps.
 *
 * Copyright (C) 2010 Prodevelop.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,USA.
 *
 * For more information, contact:
 *
 *   Prodevelop, S.L.
 *   Pza. Don Juan de Villarrasa, 14 - 5
 *   46001 Valencia
 *   Spain
 *
 *   +34 963 510 612
 *   +34 963 510 968
 *   prode@prodevelop.es
 *   http://www.prodevelop.es
 *
 *   gvSIG Mini has been partially funded by IMPIVA (Instituto de la Peque�a y
 *   Mediana Empresa de la Comunidad Valenciana) &
 *   European Union FEDER funds.
 *   
 *   2010.
 *   author Alberto Romeu aromeu@prodevelop.es
 *   
 */

package es.prodevelop.android.spatialindex.quadtree.provide.perst;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Hashtable;
import java.util.Iterator;

import org.garret.perst.Query;
import org.garret.perst.RectangleR2;
import org.garret.perst.SpatialIndexR2;
import org.garret.perst.fulltext.FullTextSearchResult;

import es.prodevelop.android.spatialindex.poi.Metadata;
import es.prodevelop.android.spatialindex.poi.OsmPOIStreet;
import es.prodevelop.android.spatialindex.poi.POI;
import es.prodevelop.android.spatialindex.poi.POICategories;
import es.prodevelop.android.spatialindex.quadtree.persist.perst.OSMPoiSpatialIndexRoot;
import es.prodevelop.android.spatialindex.quadtree.persist.perst.PerstDBHelper;
import es.prodevelop.android.spatialindex.quadtree.persist.perst.SpatialIndexRoot;
import es.prodevelop.android.spatialindex.quadtree.provide.FullTextSearchListener;
import es.prodevelop.gvsig.mini.exceptions.BaseException;
import es.prodevelop.gvsig.mini.geom.Extent;
import es.prodevelop.gvsig.mini.utiles.Cancellable;

/**
 * A provider of OsmPOI instances. It looks for OsmPOis on the SpatialIndexR2 on
 * the Perst database. @see {@link SpatialIndexRoot#getPoiIndex()}
 * 
 * @author aromeu
 * 
 */
public class PerstOsmPOIProvider implements PerstQuadtreeProvider {

	// default is 1319
	// Storage.setProperty("perst.object.cache.init.size", new Integer(?))
	private String databasePath;
	private PerstDBHelper helper;
	private Hashtable categoriesOffset = new Hashtable();
	private FullTextSearchListener listener;

	public PerstOsmPOIProvider(String databasePath) {
		this.databasePath = databasePath;
		helper = new PerstDBHelper();
		helper.openDatabase(this.databasePath, PerstDBHelper.DEFAULT_PAGE_POOL);
	}

	public Collection getPOIs(Extent boundingBox, int zoomLevel,
			Cancellable cancellable) throws BaseException {
		final RectangleR2 rect = new RectangleR2(boundingBox.getMinY(),
				boundingBox.getMinX(), boundingBox.getMaxY(),
				boundingBox.getMaxX());
		return getIndex().getList(rect);
	}

	/**
	 * @return the helper
	 */
	public PerstDBHelper getHelper() {
		return helper;
	}

	/**
	 * @param helper
	 *            the helper to set
	 */
	public void setHelper(PerstDBHelper helper) {
		this.helper = helper;
	}

	public ArrayList fullTextSearch(String text) {
		ArrayList list = new ArrayList();
		SpatialIndexRoot root = ((SpatialIndexRoot) getHelper().getRoot());

		FullTextSearchResult result = root.getFullTextIndex().search(text,
				SpatialIndexRoot.DEFAULT_LANGUAGE,
				SpatialIndexRoot.DEFAULT_MAX_RESULTS,
				SpatialIndexRoot.DEFAULT_MAX_TIME);

		final int size = result.hits.length;
		for (int i = 0; i < size; i++) {
			list.add(result.hits[i].getDocument());
		}

		if (listener != null)
			listener.onSearchResults(text, list);
		return list;
	}

	SpatialIndexR2 getIndex() {
		return ((SpatialIndexRoot) getHelper().getRoot()).getPoiIndex();
	}

	public Collection getPOIsByName(String name, boolean exactMatch,
			Cancellable cancellable) throws BaseException {
		if (exactMatch)
			return searchExactPOIByName(name);
		else
			return searchPOIByName(name);
	}

	public Collection searchExactPOIByName(String name) {
		return null;
	}

	public Collection searchPOIByName(String name) {
		return null;
	}

	public Collection getPOIs(Extent boundingBox, double res, int zoomLevel,
			Cancellable cancellable) throws BaseException {
		return this.getPOIs(boundingBox, zoomLevel, cancellable);
	}

	public Iterator getStreetsByName(String name, boolean exactMatch) {
		SpatialIndexRoot root = ((SpatialIndexRoot) this.getHelper().getRoot());
		Query q1 = this.getHelper().createQuery();
		q1.addIndex("description", root.getStreetIndex());
		if (exactMatch) {
			q1.prepare(OsmPOIStreet.class, "description = ?");
			q1.setParameter(1, name);
		} else {
			q1.prepare(OsmPOIStreet.class, "description like ?");
			q1.setParameter(1, "%" + name + "%");
		}

		return q1.execute(root.getStreetIndex().iterator());
	}

	public Collection getStreetsByPrefixName(String prefix) {
		SpatialIndexRoot root = ((SpatialIndexRoot) this.getHelper().getRoot());
		return root.getStreetIndex().getPrefixList(prefix);
	}

	public Metadata getPOIMetadata() {
		SpatialIndexRoot root = ((SpatialIndexRoot) this.getHelper().getRoot());
		return (Metadata) root.getMetadataIndex().get(
				SpatialIndexRoot.METADATA_POI_KEY);
	}

	public Metadata getStreetMetadata() {
		SpatialIndexRoot root = ((SpatialIndexRoot) this.getHelper().getRoot());
		return (Metadata) root.getMetadataIndex().get(
				SpatialIndexRoot.METADATA_STREET_KEY);
	}

	public POI getAt(String category, int index) {
		if (category.compareTo(POICategories.STREETS) == 0) {
			return (POI) ((SpatialIndexRoot) getHelper().getRoot())
					.getStreetIndex().getAt(index);
		} else {
			int offset = -1;
			if (categoriesOffset.containsKey(category))
				offset = ((Integer) categoriesOffset.get(category)).intValue();
			else {
				offset = getPOIMetadata().getCategoryOffset(category);
				categoriesOffset.put(category, new Integer(offset));
			}

			return (POI) ((OSMPoiSpatialIndexRoot) getHelper().getRoot())
					.getCatSubCategoryIndex().getAt(index + offset);
		}
	}

	public String getDatabasePath() {
		return this.databasePath;
	}

	public FullTextSearchListener getListener() {
		return listener;
	}

	public void setListener(FullTextSearchListener listener) {
		this.listener = listener;
	}
}
